open Ldap_types
open Ldap_ooclient
open Ldap_error
open Tdir_types
open Str
open Syslog

let add_res ?(referral=None) ?(code=`SUCCESS) ?(mdn="") ?(err="success") msg = 
  {messageID=msg.messageID;
   protocolOp=Add_response {result_code=code;
			    matched_dn=mdn;
			    error_message=err;
			    ldap_referral=referral};
   controls=None}

let error message msg = add_res message ~code:`OTHER ~err:msg

let generate_objectcategory cfg original_entry e = 
  match Tdir_utils.object_type cfg original_entry with
      `User -> 
	["cn=person,cn=schema,cn=configuration," ^ !cfg.cf_real_replica_base]
    | `Group ->
	["cn=group,cn=schema,cn=configuration," ^ !cfg.cf_real_replica_base]
    | `Unrecognized ->
	raise (Generation_failed (Generator_error "unrecognized object type"))

let generate_grouptype e = ["-2147483646"]

let generate_samaccountname cfg original_entry e =
  try
    [try List.hd (original_entry#get_value !cfg.cf_imaginary_replica_rdn)
     with Not_found -> List.hd (original_entry#get_value !cfg.cf_real_replica_rdn)]
  with Not_found -> 
    raise (Generation_failed 
	     (Generator_error 
		(Printf.sprintf 
		   "cannot find a suitable attribute to use as samaccountname, tried: %s, and %s, but neither exist"
		   !cfg.cf_imaginary_replica_rdn
		   !cfg.cf_real_replica_rdn)))

let map_attrs cfg schema original_entry adentry = 
  let attrs = original_entry#attributes in
  let mapping attrs matchre = 
    List.find (fun attr -> Pcre.pmatch ~rex:matchre attr) attrs
  in
    List.iter
      (fun {attribute_to_match=matchre;attribute_to_emit=mappingname} ->
	 try
	   let attr = mapping attrs matchre in
	     adentry#replace 
	       [(mappingname, 
		 if (try (* if the attr doesn't exist it will get removed anyway *)
		       (getAttr 
			  schema 
			  (Ldap_schemaparser.Lcstring.of_string mappingname)
		       ).Ldap_schemaparser.at_single_value
		     with _ -> false)
		 then
		   [List.hd (original_entry#get_value attr)]
		 else
		    original_entry#get_value attr)]
	 with Not_found -> ()) (* the entry doesn't have an attribute which matches *)
      !cfg.cf_attribute_mappings

let map_ids cfg original_entry account = 
  List.fold_left
    (fun lst membership -> 
       try 
	 let group_dn = (Tdir_utils.find_object_dn cfg membership) in
	   (group_dn, `ADD, !cfg.cf_real_group_join_attr, [account#dn]) :: lst
       with Not_found | Tdir_utils.Util_failure _ -> lst)
    []
    (try original_entry#get_value !cfg.cf_imaginary_group_join_attr
     with Not_found -> [])

let add_ids_to_group cfg connection original_obj account = 
  match map_ids cfg original_obj account with (* group id mapping *)
      [] -> ()
    | mod_ops ->
	(List.iter
	   (fun (dn, op, attr, vals) ->
	      try
		if (List.for_all 
		      (fun rex -> Pcre.pmatch ~rex:rex dn) 
		      !cfg.cf_target_group_must_match) &&
				        (not (List.exists
						(fun rex -> Pcre.pmatch ~rex:rex dn)
						!cfg.cf_target_group_must_not_match))
		then
		  connection.con_adcon#modify dn [(op, attr, vals)]
	      with 
		  (LDAP_Failure _) as exn ->
		    syslog !cfg.cf_syslog `LOG_ERR
		      (ldap_strerror ("member add failed: " ^ dn) exn)
		| exn -> 
		    syslog !cfg.cf_syslog `LOG_ERR
		      ("member add failed: " ^ dn ^ ": " ^ (Printexc.to_string exn)))
	   mod_ops)

let add_group cfg generators message connection schema group = 
  let account = (let a = new ldapaccount schema generators (Hashtbl.create 1) in
		   a#of_entry group;a)
  in
    account#set_changetype `ADD;
    (if !cfg.cf_replica_type = Active_directory then      
       account#add [("objectclass", ["group"])]);
    account#generate;
    account#set_dn (Tdir_utils.translate_dn cfg account#dn);
    (if !cfg.cf_debugging then account#print);    
    (if !cfg.cf_replica_type = Active_directory then
       (match
	  (List.rev_map
	     (fun (dn, op, attr, vals) -> (op, attr, vals))
	     (map_ids cfg group account))
	with
	    [] -> ()
	  | mods -> account#modify mods);
     connection.con_adcon#add (account :> ldapentry));
    add_res message

let add_user cfg generators message connection schema user = 
  let account = 
    let a = new ldapaccount schema generators (Hashtbl.create 1) in
      a#of_entry ~scflavor:!cfg.cf_schemacheck_flavor user;
      a
  in
    account#set_changetype `ADD;
    (if !cfg.cf_replica_type = Active_directory then
       account#add [("objectclass", ["user"])]);
    account#set_dn (Tdir_utils.translate_dn cfg account#dn);
    account#replace [(!cfg.cf_real_replica_rdn, [Tdir_utils.rdn_value account#dn])];
    account#generate;
    (* special handling for the password in active directory *)
    (if !cfg.cf_replica_type = Active_directory then
       (if user#exists "userpassword" then
	  let passwords =
	    List.filter
	      (fun pwd ->
		 (not
		    (List.exists
		       (fun rx -> Pcre.pmatch ~rex:rx pwd)
		       !cfg.cf_userpassword_filter)))
	      (user#get_value "userpassword")			    
	  in
	    (if not (passwords = []) then
	       account#add
		 [("unicodepwd",
		   [Tdir_utils.mkunicodepwd 
		      (List.hd (user#get_value "userpassword"))]);
		  ("useraccountcontrol", ["66080"])]
	     else
	       account#add [("useraccountcontrol", ["66082"])])
	else
	  account#add [("useraccountcontrol", ["66082"])]));
    map_attrs cfg schema user account; (* perform attribute mappings *)
    (if !cfg.cf_debugging then account#print);
    connection.con_adcon#add (account :> ldapentry);
    (if !cfg.cf_replica_type = Active_directory then
       add_ids_to_group cfg connection user account);
    add_res message
      
let add cfg {connections=connections;ad_schema=schema} conn_id message =
  match message with
      {controls=Some cntrls} when List.exists (fun cntrl -> cntrl.criticality) cntrls ->
	add_res message ~code:`UNAVAILABLE_CRITICAL_EXTENSION ~err:""
    | {messageID=id;
       protocolOp=Add_request e} ->
	if (List.exists (* make sure the dn matches criteria *)
	      (fun rx -> Pcre.pmatch ~rex:rx e.sr_dn) 
	      !cfg.cf_dn_must_not_match) || 
	  (not
	     (List.exists
		(fun rx -> Pcre.pmatch ~rex:rx e.sr_dn)
		!cfg.cf_dn_must_match)) ||
	      (not (Tdir_utils.match_object cfg e.sr_dn)) (* make sure the object matches criteria *)
	then
	  (syslog !cfg.cf_syslog `LOG_DEBUG "dn does not meet criteria, not doing add";
	   add_res message)
	else
	  (try
	     let connection = Tdir_utils.connection cfg connections conn_id in
	     let entry = to_entry (`Entry e) in
	     let generators =
	       let tbl = Hashtbl.create 3 in
		 if !cfg.cf_replica_type = Active_directory then begin
		   Hashtbl.add tbl 
		     "objectcategory" 
		     {gen_name="objectcategory";
		      required=[];
		      genfun=(generate_objectcategory cfg entry)};
		   Hashtbl.add tbl 
		     "samaccountname" 
		     {gen_name="samaccountname";
		      required=[];
		      genfun=(generate_samaccountname cfg entry)};
		   Hashtbl.add tbl
		     "grouptype"
		     {gen_name="grouptype";
		      required=[];
		      genfun=generate_grouptype}
		 end;
		 tbl
	     in
	       (match Tdir_utils.object_type cfg entry with
		    `User -> add_user cfg generators message connection schema entry
		  | `Group -> add_group cfg generators message connection schema entry
		  | `Unrecognized -> error message "unrecognized object type")
	   with
	       LDAP_Failure (code, err, {ext_matched_dn=mdn;ext_referral=refer}) ->	     
		 add_res message ~referral:refer ~code:code ~err:err ~mdn:mdn
	     | LDAP_Decoder e -> error message ("AD Protocol error: " ^ e)
	     | Invalid_objectclass e -> error message ("Invalid objectclass error: " ^ e)
	     | Invalid_attribute e -> error message ("Invalid attribute error: " ^ e)
	     | Single_value e -> error message ("The attribute is single value: " ^ e)
	     | No_generator e -> error message ("No generator for the attribute: " ^ e)
	     | Generation_failed (Missing_required missing) ->
		 error message
		   ("Required attributes are missing: " ^ 
		      (List.fold_left
			 (fun s item -> s ^ ":" ^ item)
			 "" missing))
	     | Generation_failed (Generator_error e) -> error message ("generator error: " ^ e)
	     | Generator_dep_unsatisfiable (e1, e2) -> 
		 error message ("Unsatisfiable generator dependancy: " ^ e1 ^ ":" ^ e2)
	     | Cannot_sort_dependancies deplst -> 
		 error message 
		   ("Cannot Sort Dependancies: " ^ 
		      (List.fold_left
			 (fun s item -> s ^ ":" ^ item)
			 "" deplst))
	     | Tdir_utils.Util_failure msg -> 
		 add_res ~code:`INVALID_DN_SYNTAX ~err:msg message
	     | exn -> 
		 add_res ~code:`OTHER 
		   ~err:("unhandled exception: " ^ (Printexc.to_string exn)) message)
    | _ -> add_res message ~code:`PROTOCOL_ERROR ~err:""
