open Tdir_types
open Inifiles
open Syslog
open Printf
open Ldap_ooclient
open Ldap_types
open Ldap_error

let read_dn_special_cases cfg =
  let to_match = 
    List.map 
      (fun rex -> Pcre.regexp ~flags:[`CASELESS] rex)
      (try cfg#getaval "dn-translation" "match-special-case" with _ -> [])
  in
  let to_emit = try cfg#getaval "dn-translation" "emit" with _ -> [] in
    if (List.length to_match) = (List.length to_emit) then
      (List.map2
	 (fun matc emit -> {dn_to_match=matc;dn_to_emit=emit})
	 to_match
	 to_emit)
    else 
      failwith 
	("invalid dn_translations, you must have the same number of emit terms" ^ 
	   " as match_special_case terms")

let read_attribute_mappings cfg =
  let to_match = 
    List.map
      (Pcre.regexp ~flags:[`CASELESS])
      (try cfg#getaval "mappings" "match" with _ -> [])
  in
  let to_emit = (try cfg#getaval "mappings" "emit" with _ -> []) in
    if (List.length to_match) = (List.length to_emit) then
      (List.map2
	 (fun matc emit -> {attribute_to_match=matc;attribute_to_emit=emit})
	 to_match
	 to_emit)
    else 
      failwith
	("invalid attribute_mappings, you must have the same number of emit terms " ^
	   "as match terms")

let read_add_musts cfg =
  try
    List.rev_map
      (fun amust -> 
	 match Str.split (Str.regexp "  *") amust with
	     [] | [_] -> failwith "malformed add-must"
	   | lst -> (List.hd lst, List.tl lst))
      (cfg#getaval "attributes" "add-must")
  with Invalid_element _ | Invalid_section _ -> []

let read_group_join_type cfg =
  try
    match cfg#getval "object-types" "group-join-type" with
	"user-id-on-group" -> Uid_on_group
      | "group-id-on-user" -> Gid_on_user
      | _ -> failwith ("invalid join type. Join type must be one of " ^ 
			 "user-id-on-group or group-id-on-user")
  with Invalid_section _ | Invalid_element _ -> Uid_on_group

let getconfig path = 
  let cfg = new inifile path in
  let filter_directory = 
    (try
       let ld = new ldapcon [cfg#getval "replication-criteria" "filter-directory"] in
	 ld#bind 
	   (try 
	      cfg#getval 
		"replication-criteria" 
		"filter-directory-binddn" 
	    with _ -> "")
	   ~cred:(try 
		    cfg#getval 
		      "replication-criteria"
		      "filter-directory-bindpw" 
		  with _ -> "");
	 Some ld
     with 
	 LDAP_Failure (c, e, ext) -> 
	   failwith 
	     (ldap_strerror 
		"cannot connect to filter directory"
		(LDAP_Failure (c, e, ext)))
       | _ -> None)
  in
    {cf_read_con=(let ldap = 
		    new ldapcon 
		      ~connect_timeout:(int_of_string
					  (try 
					     cfg#getval 
					       "replica-servers" 
					       "connect-timeout"
					   with _ -> "60"))
		      (cfg#getaval "replica-servers" "server")
		  in
		    ldap#bind 
		      (try cfg#getval "read" "bind-dn" with _ -> "") 
		      ~cred:(try cfg#getval "read" "bind-cred" with _ -> "");
		    ldap);
     cf_replica_type=
	(match cfg#getval "replica-servers" "replica-type" with
	     "active-directory" -> Active_directory
	   | "e-directory" -> E_directory
	   | "generic-directory" -> Standard_directory
	   | _ -> raise (Invalid_argument
			   ("replica-type must be one of active-directory, " ^ 
			      "e-directory, generic-directory")));
     cf_schemacheck_flavor=
	(try 
	   match cfg#getval "replica-servers" "schemacheck-flavor" with
	       "optimistic" -> Optimistic
	     | "pessimistic" -> Pessimistic
	     | _ -> failwith "invalid schema check flavor, use optimistic or pessimistic"
	 with Invalid_section _ | Invalid_element _ -> Pessimistic);
     cf_port=(int_of_string (try cfg#getval "server" "listenport" with _ -> "389"));
     cf_replica_hosts=(cfg#getaval "replica-servers" "server");
     cf_replica_connect_timeout=(int_of_string 
			      (try cfg#getval "replica-servers" "connect-timeout"
			       with _ -> "60"));
     cf_real_replica_base=(cfg#getval "dn-translation" "real-replica-base");
     cf_imaginary_replica_base=(cfg#getval "dn-translation" "imaginary-replica-base");
     cf_real_replica_rdn=(try cfg#getval "dn-translation" "real-replica-rdn" with _ -> "cn");
     cf_imaginary_replica_rdn=
       (try cfg#getval "dn-translation" "imaginary-ad-rdn" with _ -> "uid");
     cf_dn_special_cases=(read_dn_special_cases cfg);
     cf_group_join_type=(read_group_join_type cfg);
     cf_imaginary_group_join_attr=(try cfg#getval "object-types" "imaginary-group-join-attr"
				   with _ -> "memberuid");
     cf_real_group_join_attr=(try cfg#getval "object-types" "real-group-join-attr"
			      with _ -> "memberOf");
     cf_user_object_match=(try cfg#getval "object-types" "user-match"
			   with _ -> "person");
     cf_group_object_match=(try cfg#getval "object-types" "group-match"
			    with _ -> "posixgroup");
     cf_attributes_add_musts=(read_add_musts cfg);
     cf_attributes_ignore_required=(try cfg#getaval "attributes" "ignore-required" 
				    with _ -> []); 
     cf_attributes_single_value=(try cfg#getaval "attributes" "single-value" 
				 with _ -> []);
     cf_attributes_strip=(try cfg#getaval "attributes" "strip" 
			  with _ -> []);
     cf_objectclasses_strip=(try cfg#getaval "attributes" "strip-objectclass"
			     with _ -> []);
     cf_attribute_mappings=(read_attribute_mappings cfg);
     cf_userpassword_filter=
	(List.rev_map 
	   (Pcre.regexp ~flags:[`CASELESS])
	   (try cfg#getaval "password-replication-criteria" "password-must-not-match"
	    with _ -> []));
     cf_dn_must_match=
	(List.rev_map
	   (Pcre.regexp ~flags:[`CASELESS])
	   (try cfg#getaval "replication-criteria" "dn-must-match"
	    with _ -> ["^.*$"]));
     cf_dn_must_not_match=
	(List.rev_map
	   (Pcre.regexp ~flags:[`CASELESS])
	   (try cfg#getaval "replication-criteria" "dn-must-not-match"
	    with _ -> []));
     cf_target_group_must_match=
	(List.rev_map
	   (Pcre.regexp ~flags:[`CASELESS])
	   (try cfg#getaval "replication-criteria" "target-group-must-match"
	    with _ -> []));
     cf_target_group_must_not_match=
	(List.rev_map
	   (Pcre.regexp ~flags:[`CASELESS])
	   (try cfg#getaval "replication-criteria" "target-group-must-not-match"
	    with _ -> []));
     cf_filter_directory=filter_directory;
     cf_object_must_match=
	(try
	   let filter = 
	     Ldap_filter.to_string (* parse the filter in advance *)
	       (Ldap_filter.of_string 
		  (cfg#getval
		     "replication-criteria"
		     "object-must-match"))
	   in
	     if filter_directory = None then
	       failwith "if you specify object-must-match, then you must specify filter-directory"
	     else
	       Some filter
	 with 
	     Ldap_filter.Invalid_filter (pos, error) -> 
	       failwith (sprintf "object-must-match, bad search filter: %s, at %d" error pos)	       
	   | _ -> None);
     cf_find_object_dn_cache_timeout=
	(try Int32.of_string (cfg#getval "server" "find-object-dn-cache-timeout")
	 with exn -> 
	   if exn = Failure "int_of_string" then 
	     failwith "find-object-dn-cache-timeout must be a number"
	   else 30l);
     cf_debugging=(match try cfg#getval "server" "debugging" with _ -> "false" with
		       "true" -> true
		     | "false" -> false
		     | _ -> raise (Invalid_argument "debugging must be either true or false"));
     cf_syslog=(openlog 
		  ~facility:(facility_of_string 
			       (try cfg#getval "server" "syslogfacility"
				with _ -> "daemon"))
		  ~flags:[`LOG_PID] "tdir")}
