open Ldap_types
open Ldap_ooclient
open Ldap_schemaparser
open Ldap_error
open Tdir_types
open CharEncoding
open Syslog
open Printf
open Ldap_error

exception Util_failure of string

let lc = String.lowercase

(* work around microsoft's broken concept of unicode.
   They don't want a header with byte ordering, they assume it is big endian.
   And the thing must be in quotes...what a mess. *)
let mkunicodepwd pwd = 
  let upwd = recode_string ~in_enc:ascii ~out_enc:utf16be pwd in
    (* strip the unicode header, ad doesn't like it *)  
  let stripedupwd = String.sub upwd 2 ((String.length upwd) - 2) in
    (* The password must be in quotes. But the quotes are not part of the password *)
    "\"" ^ stripedupwd ^ "\000\"\000"

let connection cfg connections conn_id =
  try Hashtbl.find connections conn_id
  with Not_found -> (* new connection *)
    let c = {con_bound=false;
	     con_adcon=(new ldapcon 
			  ~connect_timeout:!cfg.cf_replica_connect_timeout
			  !cfg.cf_replica_hosts)}
    in
      Hashtbl.add connections conn_id c;c

let object_type cfg e = 
  if (List.exists 
	(fun oc -> lc !cfg.cf_user_object_match = lc oc)
	(e#get_value "objectclass")) 
  then
    `User
  else if (List.exists 
	     (fun oc -> lc !cfg.cf_group_object_match = lc oc)
	     (e#get_value "objectclass"))
  then
    `Group
  else
    `Unrecognized
      
(* this is used for group replication to active directory *)
let find_object_dn_cache = Hashtbl.create 10
let find_object_dn_cache_last_sweep = ref 0l
let sweep_fdn_cache cfg = 
  let now = Int32.of_float (Unix.gettimeofday ()) in
  let dead = 
    Hashtbl.fold
      (fun k {fdn_time=ctime} l -> 
	 if (Int32.sub now ctime) > !cfg.cf_find_object_dn_cache_timeout then
	   k :: l
	 else l)
      find_object_dn_cache
      []
  in
    List.iter (Hashtbl.remove find_object_dn_cache) dead
let find_object_dn cfg identifier = 
  try
    let {fdn_time=ctime;fdn_dn=dn} = Hashtbl.find find_object_dn_cache identifier in
    let now = Int32.of_float (Unix.gettimeofday ()) in
      if (Int32.sub now !find_object_dn_cache_last_sweep) > 30l then begin	
	sweep_fdn_cache cfg;
	find_object_dn_cache_last_sweep := now
      end;
      if (Int32.sub now ctime) < !cfg.cf_find_object_dn_cache_timeout then
	(* its alive *)
	dn
      else 	
	(* its dead *)
	raise Not_found
  with Not_found ->
    (try
       if identifier = "" then raise (Util_failure "invalid_argument");
       let candidates =
	 List.filter
	   (fun e -> not (e#dn = "referral"))
	   (!cfg.cf_read_con#search 
	      ~base:!cfg.cf_real_replica_base 
	      (!cfg.cf_real_replica_rdn ^ "=" ^ identifier))
       in
	 if List.length candidates = 1 then begin
	   Hashtbl.replace (* add to the cache *)
	     find_object_dn_cache identifier 
	     {fdn_dn=(List.hd candidates)#dn;
	      fdn_time=(Int32.of_float (Unix.gettimeofday ()))};
	   (List.hd candidates)#dn
	 end else raise (Util_failure "find_group_dn: exactly one not found")
     with (LDAP_Failure (c, e, ext)) as exn -> 
       syslog !cfg.cf_syslog `LOG_ERR (ldap_strerror "find_group_dn" exn);
       raise (Util_failure "find_group_dn"))
	
let match_object cfg dn = 
  match !cfg.cf_filter_directory with
      Some ldap ->
	(match !cfg.cf_object_must_match with
	     Some filter ->
	       (try
		  let res = ldap#search ~attrs:["dn"] ~base:dn filter in
		    if List.length res = 1 then true
		    else begin
		      syslog !cfg.cf_syslog `LOG_ERR
			(sprintf "object: %s does not match required filter, %d objects returned"
			   dn (List.length res));
		      false
		   end
		with LDAP_Failure (c, e, ext) ->
		  syslog !cfg.cf_syslog `LOG_ERR
		    (ldap_strerror
		       "failed to search for object match, returning false"
		       (LDAP_Failure (c, e, ext)));
		  false)
	   | None -> true)
    | None -> true

let dn_normalize dn = String.lowercase dn

let seprex = Str.regexp " *, *"
let explode_dn dn = Str.split seprex dn

let attrvalrex = Str.regexp " *= *"
let rdn_value dn =
  try
    (List.nth
       (Str.split attrvalrex
	  (List.hd
	     (explode_dn dn)))
       1)
  with
      Failure "hd"
    | Failure "nth" -> raise (Util_failure "rdn_value: invalid dn")

let reconstitute_dn parts = 
  let rec reconstitute_dn' dn parts =
    match parts with
	car :: cdr -> (* honor thy ancestors *)
	  reconstitute_dn' (dn ^ "," ^ car) cdr
      | [] -> dn
  in
    match parts with
	hd :: tl -> reconstitute_dn' hd tl
      | [] -> ""

let translate_dn cfg dn = 
  let special_cases =
    List.filter
      (fun {dn_to_match=rx} -> Pcre.pmatch ~rex:rx dn)
      !cfg.cf_dn_special_cases
  in
    match special_cases with
	[] ->
	  let rec remove_ibase edn1 edn2 =
	    match edn1 with
		hd :: tl -> 
		  (match edn2 with
		       hd1 :: tl1 ->
			 if hd = hd1 then
			   remove_ibase tl tl1
			 else edn2
		     | [] -> [])
	      | [] -> edn2
	  in
	  let patch_rdn cfg rdn =
	    match Str.split (Str.regexp " *= *") rdn with
		[rdnattr;rdnval] -> 
		  if (lc rdnattr) = (lc !cfg.cf_imaginary_replica_rdn) then
		    (* replace the imaginary with the real *)
		    !cfg.cf_real_replica_rdn ^ "=" ^ rdnval
		  else if (lc rdnattr) = (lc !cfg.cf_real_replica_rdn) then
		    (* already using the real *)
		    rdn
		  else
		    (* not using either, just pass it through *)
		    rdn
	      | _ -> raise (Util_failure "failed to tokenize rdn")
	  in
	    (reconstitute_dn
	       (match
		  (List.rev
		     (List.append
			(List.rev 
			   (explode_dn 
			      (dn_normalize 
				 !cfg.cf_real_replica_base)))
			(remove_ibase
			   (List.rev
			      (explode_dn 
				 (dn_normalize !cfg.cf_imaginary_replica_base)))
			   (List.rev 
			      (explode_dn 
				 (dn_normalize dn))))))
		with
		    hd :: tl -> (patch_rdn cfg hd) :: tl
		  | [] -> raise (Util_failure "translated result is the empty list!")))
      | [{dn_to_emit=emit}] -> emit
      | _ -> raise (Util_failure "translate_dn more than one special case matches")

let alter_schema cfg (schema:Ldap_schemaparser.schema) = 
  let equate_attr a1 a2 = (attrToOid schema a1) = (attrToOid schema a2) in
  let find_ocs_with_must_or_may schema mustormay attr = 
    Hashtbl.fold
      (fun k v l ->
	 if (List.exists 
	       (fun a -> 
		  try equate_attr (Lcstring.of_string attr) a
		  with Invalid_attribute a -> 
		    syslog !cfg.cf_syslog `LOG_ERR 
		      ("find_ocs_with_must_or_may: warning " ^
			 "objectclass: " ^ (Lcstring.to_string k) ^
			 " defines an attribute named: " ^ a ^
			 " however the schema does not define " ^ a);
		    (* we know this isn't the one because ours is at least
		       defined in the schema! *)
		    false)
	       (match mustormay with
		    `Must -> v.oc_must
		  | `May -> v.oc_may))
	 then
	   (k, v) :: l
	 else l)
      schema.objectclasses
      []
  in
  let strip_musts_or_mays schema mustormay attrs =
      (List.iter
	 (fun tostrip ->
	    (List.iter
	       (fun (name, oc) ->
		  let newoc = 
		    (match mustormay with
			 `Must ->
			   {oc with oc_must=
			      (List.filter
				 (fun m -> not (equate_attr (Lcstring.of_string tostrip) m))
				 oc.oc_must)}
		       | `May ->
			   {oc with oc_may=
			      (List.filter
				 (fun m -> not (equate_attr (Lcstring.of_string tostrip) m))
				 oc.oc_may)})
		  in
		    Hashtbl.replace schema.objectclasses name newoc;
		    Hashtbl.replace schema.objectclasses_byoid (ocToOid schema name) newoc)
	       (find_ocs_with_must_or_may schema mustormay tostrip)))
	 attrs)
  in
    (* despite what the schema says, the following attributes are considered
       SINGLE-VALUE by AD. They may be not be correctly marked as such, so we
       mark them now *)
  let make_single_value (schema:schema) attrs =
    List.iter
      (fun attr ->
	 try
	   let newattr = 
	     {(Hashtbl.find schema.attributes_byoid 
		 (attrToOid schema (Lcstring.of_string attr)))
	      with
		at_single_value=true}
	   in
	     Hashtbl.replace schema.attributes (Lcstring.of_string attr) newattr;
	     Hashtbl.replace schema.attributes_byoid 
	       (attrToOid schema (Lcstring.of_string attr)) newattr
	 with Not_found | Invalid_attribute _ ->
	   syslog !cfg.cf_syslog `LOG_ERR
	     (sprintf "make_single_value: warning, attribute: %s does not exist\n" attr))
      attrs
  in
  let remove_attrs schema attrs = 
    List.iter
      (fun attr ->
	 try
	   if (Hashtbl.mem schema.attributes_byoid
		 (attrToOid schema (Lcstring.of_string attr)))
	   then
	     (* remove all record of the attribute's existance *)
	     (strip_musts_or_mays schema `May [attr];
	      strip_musts_or_mays schema `Must [attr];
	      Hashtbl.remove schema.attributes_byoid 
		(attrToOid schema (Lcstring.of_string attr));
	      Hashtbl.remove schema.attributes (Lcstring.of_string attr))
	 with Not_found | Invalid_attribute _ ->
	   syslog !cfg.cf_syslog `LOG_ERR
	     (sprintf "remove_attrs: warning, attribute: %s does not exist\n" attr))
      attrs
  in
  let add_musts schema musts = 
    List.iter
      (fun (oc, vals) ->
	 try
	   (Hashtbl.replace schema.objectclasses (Lcstring.of_string oc)
	      (let theoc = getOc schema (Lcstring.of_string oc) in
		 {theoc with oc_must=(List.rev_append theoc.oc_must 
					(List.rev_map Lcstring.of_string vals))}))
	 with Not_found | Invalid_objectclass _ -> 
	   syslog !cfg.cf_syslog `LOG_ERR
	     (sprintf "add_musts: warning, objectclass: %s doesn't exist\n" oc))
      musts
  in
  let rec strip_objectclass schema oc = 
    let find_ocs_with_superior schema sup = 
      Hashtbl.fold
	(fun k ({oc_sup=mysup} as v) l -> 
	   if List.mem sup mysup then
	     v :: l
	   else l)
	schema.objectclasses []
    in
    let theoc = getOc schema (Lcstring.of_string oc) in
      (List.iter
	 (fun ocname ->
	    let ocname = Lcstring.of_string ocname in
	      (List.iter 
		 (fun {oc_name=names} -> 
		    (List.iter 
		       (fun name -> 
			  try strip_objectclass schema name 
			  with Invalid_objectclass _ -> ()) 
		       names))
		 (find_ocs_with_superior schema ocname));
	      Hashtbl.remove schema.objectclasses ocname)
	 theoc.oc_name);
      Hashtbl.remove schema.objectclasses_byoid theoc.oc_oid
  in
  let strip_objectclasses schema ocs = 
    List.iter
      (fun oc ->
	 try strip_objectclass schema oc
	 with Invalid_objectclass c ->
	   syslog !cfg.cf_syslog `LOG_ERR (sprintf "objectclass does not exist: %s" c))
      ocs
  in
    strip_objectclasses schema !cfg.cf_objectclasses_strip;
    strip_musts_or_mays schema `Must !cfg.cf_attributes_ignore_required;
    make_single_value schema !cfg.cf_attributes_single_value;
    remove_attrs schema !cfg.cf_attributes_strip;
    add_musts schema !cfg.cf_attributes_add_musts
